import contextvars
import os
import uuid
from datetime import UTC, datetime

from pydantic import BaseModel, Field, model_validator

from alphatrion.metadata.sql_models import COMPLETED_STATUS, TrialStatus
from alphatrion.run.run import Run
from alphatrion.runtime.runtime import global_runtime
from alphatrion.utils.context import Context

# Used in log/log.py to log params/metrics
current_trial_id = contextvars.ContextVar("current_trial_id", default=None)


class CheckpointConfig(BaseModel):
    """Configuration for a checkpoint."""

    enabled: bool = Field(
        default=False,
        description="Whether to enable checkpointing. \
            Default is False.",
    )
    # save_every_n_seconds: int | None = Field(
    #     default=None,
    #     description="Interval in seconds to save checkpoints. \
    #         Default is None.",
    # )
    # save_every_n_steps: int | None = Field(
    #     default=None,
    #     description="Interval in steps to save checkpoints. \
    #         Default is None.",
    # )
    save_on_best: bool = Field(
        default=False,
        description="Once a best result is found, it will be saved. \
            The metric to monitor is specified by monitor_metric. Default is False. \
            Can be enabled together with save_every_n_steps/save_every_n_seconds.",
    )
    path: str = Field(
        default="checkpoints",
        description="The path to save checkpoints. Default is 'checkpoints'.",
    )


class TrialConfig(BaseModel):
    """Configuration for an experiment."""

    max_duration_seconds: int = Field(
        default=-1,
        description="Maximum duration in seconds for the experiment. \
        Default is -1 (no limit).",
    )
    early_stopping_runs: int = Field(
        default=-1,
        description="Number of runs with no improvement \
        after which experiment will be stopped. Default is -1 (no early stopping). \
        Count each time when calling log_metrics with the monitored metric.",
    )
    max_runs_per_trial: int = Field(
        default=-1,
        description="Maximum number of runs for each trial. \
        Default is -1 (no limit). Count by the finished runs.",
    )
    monitor_metric: str | None = Field(
        default=None,
        description="The metric to monitor together with other configurations  \
            like early_stopping_runs and save_on_best. \
            Required if save_on_best is true or early_stopping_runs > 0.",
    )
    monitor_mode: str = Field(
        default="max",
        description="The mode for monitoring the metric. Can be 'max' or 'min'. \
            Default is 'max'.",
    )
    checkpoint: CheckpointConfig = Field(
        default=CheckpointConfig(),
        description="Configuration for checkpointing.",
    )

    @model_validator(mode="after")
    def metric_must_be_valid(self):
        if self.checkpoint.save_on_best and not self.monitor_metric:
            raise ValueError(
                "monitor_metric must be specified \
                when checkpoint.save_on_best=True"
            )
        if self.early_stopping_runs > 0 and not self.monitor_metric:
            raise ValueError(
                "monitor_metric must be specified \
                when early_stopping_runs>0"
            )
        return self


class Trial:
    __slots__ = (
        "_id",
        "_exp_id",
        "_config",
        "_runtime",
        # step is used to track the round, e.g. the step in metric logging.
        "_step",
        "_context",
        "_token",
        # _meta stores the runtime meta information of the trial.
        # * best_metrics: dict of best metric values, used for checkpointing and
        #   early stopping. When the workload(e.g. Pod) restarts, the meta info
        #   will be lost and start from scratch. Then once some features like
        #   early_stopping_runs is enabled, it may lead to unexpected behaviors like
        #   never stopping because the counter is reset everytime restarted.
        #   To avoid this, you can set the restart times for the workload.
        "_meta",
        # key is run_id, value is Run instance
        "_runs",
        "_running_tasks",
        # Only work when early_stopping_runs > 0
        "_early_stopping_counter",
        # Only work when max_runs_per_trial > 0
        "_total_runs_counter",
    )

    def __init__(self, exp_id: int, config: TrialConfig | None = None):
        self._exp_id = exp_id
        self._config = config or TrialConfig()
        self._runtime = global_runtime()
        self._step = 0
        self._context = Context(
            cancel_func=self._stop,
            timeout=self._timeout(),
        )
        self._construct_meta()
        self._runs = dict()
        self._running_tasks = dict()
        self._early_stopping_counter = 0
        self._total_runs_counter = 0

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.cancel()
        if self._token:
            current_trial_id.reset(self._token)

    @property
    def id(self) -> uuid.UUID:
        return self._id

    def _construct_meta(self):
        self._meta = dict()

        if self._config.monitor_mode == "max":
            self._meta["best_metrics"] = {self._config.monitor_metric: float("-inf")}
        elif self._config.monitor_mode == "min":
            self._meta["best_metrics"] = {self._config.monitor_metric: float("inf")}
        else:
            raise ValueError(f"Invalid monitor_mode: {self._config.monitor_mode}")

    def config(self) -> TrialConfig:
        return self._config

    def should_checkpoint_on_best(self, metric_key: str, metric_value: float) -> bool:
        is_best_metric = self._save_if_best_metric(metric_key, metric_value)
        return (
            self._config.checkpoint.enabled
            and self._config.checkpoint.save_on_best
            and is_best_metric
        )

    def _save_if_best_metric(self, metric_key: str, metric_value: float) -> bool:
        """Save the metric if it is the best so far.
        Returns True if the metric is the best so far, False otherwise.
        """
        if metric_key != self._config.monitor_metric:
            return False

        best_value = self._meta["best_metrics"][metric_key]

        if self._config.monitor_mode == "max":
            if metric_value > best_value:
                self._meta["best_metrics"][metric_key] = metric_value
                return True
        elif self._config.monitor_mode == "min":
            if metric_value < best_value:
                self._meta["best_metrics"][metric_key] = metric_value
                return True
        else:
            raise ValueError(f"Invalid monitor_mode: {self._config.monitor_mode}")

        return False

    def should_early_stop(self, metric_key: str, metric_value: float) -> bool:
        if (
            self._config.early_stopping_runs <= 0
            or metric_key != self._config.monitor_metric
        ):
            return False

        best_value = self._meta["best_metrics"][metric_key]

        if self._config.monitor_mode == "max":
            if metric_value < best_value:
                self._early_stopping_counter += 1
            else:
                self._early_stopping_counter = 0
        elif self._config.monitor_mode == "min":
            if metric_value > best_value:
                self._early_stopping_counter += 1
            else:
                self._early_stopping_counter = 0
        else:
            raise ValueError(f"Invalid monitor_mode: {self._config.monitor_mode}")

        return self._early_stopping_counter >= self._config.early_stopping_runs

    def _timeout(self) -> int | None:
        timeout = self._config.max_duration_seconds
        if timeout < 0:
            return None

        # Adjust timeout based on the trial start time from environment variable,
        # this is useful when running in cloud env when the trial process may be
        # restarted.
        start_time = os.environ.get("ALPHATRION_TRIAL_START_TIME", None)
        if start_time is not None:
            elapsed = (
                datetime.now(UTC)
                - datetime.fromisoformat(start_time).replace(tzinfo=UTC)
            ).total_seconds()
            timeout -= int(elapsed)
        return timeout

    # Make sure you have termination condition, either by timeout or by calling cancel()
    # Before we have logic like once all the tasks are done, we'll call the cancel()
    # automatically, however, this is unpredictable because some tasks may be waiting
    # for external events, so we leave it to the user to decide when to stop the trial.
    async def wait(self):
        await self._context.wait()

    def cancelled(self) -> bool:
        return self._context.cancelled()

    # If the name is same in the same experiment, it will refer to the existing trial.
    def _start(
        self,
        name: str,
        description: str | None = None,
        meta: dict | None = None,
        params: dict | None = None,
    ):
        trial_obj = self._runtime._metadb.get_trial_by_name(
            trial_name=name, exp_id=self._exp_id
        )
        # FIXME: what if the existing trial is finished, will lead to confusion?
        if trial_obj:
            self._id = trial_obj.uuid
        else:
            self._id = self._runtime._metadb.create_trial(
                project_id=self._runtime._project_id,
                exp_id=self._exp_id,
                name=name,
                description=description,
                meta=meta,
                params=params,
                status=TrialStatus.RUNNING,
            )

        # We don't reset the trial id context var here, because
        # each trial runs in its own context.
        self._token = current_trial_id.set(self._id)
        self._context.start()

    # cancel function should be called manually as a pair of start
    # FIXME: watch for system signals to cancel the trial gracefully,
    # or it could lead to trial not being marked as finished.
    def cancel(self):
        self._context.cancel()

    def _stop(self):
        trial = self._runtime._metadb.get_trial(trial_id=self._id)
        if trial is not None and trial.status not in COMPLETED_STATUS:
            duration = (
                datetime.now(UTC) - trial.created_at.replace(tzinfo=UTC)
            ).total_seconds()
            self._runtime._metadb.update_trial(
                trial_id=self._id, status=TrialStatus.FINISHED, duration=duration
            )

        self._runtime.current_exp.unregister_trial(self._id)
        self._runs.clear()
        for task in self._running_tasks.values():
            task.cancel()
        self._running_tasks.clear()

    def _get_obj(self):
        return self._runtime._metadb.get_trial(trial_id=self._id)

    def increment_step(self) -> int:
        self._step += 1
        return self._step

    def start_run(self, call_func: callable) -> Run:
        """Start a new run for the trial.
        :param call_func: a callable function that returns a coroutine.
                          It must be a async and lambda function.
        :return: the Run instance."""

        run = Run(trial_id=self._id)
        task = run._start(call_func)
        if task is None:
            raise RuntimeError("Failed to start the run task.")
        self._runs[run.id] = run
        self._running_tasks[run.id] = task

        task.add_done_callback(lambda t: self._running_tasks.pop(run.id, None))
        task.add_done_callback(lambda t: self._runs.pop(run.id, None))
        if self._config.max_runs_per_trial > 0:
            task.add_done_callback(
                lambda t: (
                    setattr(self, "_total_runs_counter", self._total_runs_counter + 1),
                    self.cancel()
                    if self._total_runs_counter >= self._config.max_runs_per_trial
                    else None,
                )
            )

        return run



from abc import ABC, abstractmethod


class MetaStore(ABC):
    """Base class for all metadata storage backends."""

    @abstractmethod
    def get_project(self, project_id: str):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def create_exp(
        self,
        name: str,
        project_id: str,
        description: str | None = None,
        meta: dict | None = None,
    ) -> int:
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def delete_exp(self, exp_id: int):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def update_exp(self, exp_id: int, **kwargs):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def get_exp(self, exp_id: int):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def get_exp_by_name(self, name: str, project_id: str):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def list_exps(self, project_id: str, page: int, page_size: int):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def create_model(
        self,
        name: str,
        project_id: str,
        version: str = "latest",
        description: str | None = None,
        meta: dict | None = None,
    ) -> int:
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def update_model(self, model_id: int, **kwargs):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def get_model(self, model_id: int):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def list_models(self, project_id: str, page: int, page_size: int):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def delete_model(self, model_id: int):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def create_trial(
        self,
        project_id: str,
        experiment_id: str,
        name: str,
        description: str | None = None,
        meta: dict | None = None,
        params: dict | None = None,
    ) -> int:
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def get_trial(self, trial_id: int):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def get_trial_by_name(self, name: str, experiment_id: str):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def update_trial(self, trial_id: int, **kwargs):
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def create_run(self, project_id: str, trial_id: str) -> int:
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def create_metric(
        self,
        trial_id: str,
        name: str,
        value: float,
        step: int | None = None,
        timestamp: int | None = None,
    ) -> int:
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def list_metrics(self, trial_id: int) -> list[dict]:
        raise NotImplementedError("Subclasses must implement this method.")



import asyncio
import contextvars
import uuid

from alphatrion.runtime.runtime import global_runtime

current_run_id = contextvars.ContextVar("current_run_id", default=None)


class Run:
    def __init__(self, trial_id: uuid.UUID):
        self._runtime = global_runtime()
        self._trial_id = trial_id

    @property
    def id(self) -> uuid.UUID:
        return self._id

    def _start(self, call_func: callable) -> asyncio.Task | None:
        self._id = self._runtime._metadb.create_run(
            project_id=self._runtime._project_id, trial_id=self._trial_id
        )

        # current_run_id context var is used in tracing workflow/task decorators.
        token = current_run_id.set(self.id)
        try:
            # The created task will also inherit the current context,
            # including the current_trial_id, current_run_id context var.
            self._task = asyncio.create_task(call_func())
        finally:
            current_run_id.reset(token)

        return self._task

    async def wait(self):
        await self._task



# ruff: noqa: PLW0603
import os
import uuid

from alphatrion import consts
from alphatrion.artifact.artifact import Artifact
from alphatrion.metadata.sql import SQLStore

__RUNTIME__ = None


def init(
    project_id: uuid.UUID,
    artifact_insecure: bool = False,
    init_tables: bool = False,
):
    """
    Initialize the AlphaTrion runtime environment.

    :param project_id: the project ID to initialize the environment.
                       For testing purpose, you can use a random UUID.
    :param artifact_insecure: whether to use insecure connection to the
        artifact registry
    """
    global __RUNTIME__
    __RUNTIME__ = Runtime(
        project_id=project_id,
        artifact_insecure=artifact_insecure,
        init_tables=init_tables,
    )


def global_runtime():
    if __RUNTIME__ is None:
        raise RuntimeError("Runtime is not initialized. Call alphatrion.init() first.")
    return __RUNTIME__


# Runtime contains all kinds of clients, e.g., metadb client, artifact client, etc.
# Stateful information will also be stored here, e.g., current running experiment ID.
class Runtime:
    __slots__ = ("_project_id", "_metadb", "_artifact", "__current_exp")

    def __init__(
        self,
        project_id: str,
        artifact_insecure: bool = False,
        init_tables: bool = False,
    ):
        self._project_id = project_id
        self._metadb = SQLStore(
            os.getenv(consts.METADATA_DB_URL), init_tables=init_tables
        )
        self._artifact = Artifact(
            project_id=self._project_id, insecure=artifact_insecure
        )

    # current_exp is the current running experiment.
    @property
    def current_exp(self):
        return self.__current_exp

    @current_exp.setter
    def current_exp(self, value):
        self.__current_exp = value



from alphatrion.runtime.runtime import Runtime


class Model:
    def __init__(self, runtime: Runtime):
        self._runtime = runtime

    def create(
        self,
        name: str,
        project_id: str,
        description: str | None = None,
        meta: dict | None = None,
    ):
        return self._runtime._metadb.create_model(
            name=name,
            project_id=project_id,
            description=description,
            meta=meta,
        )

    def update(self, model_id: int, **kwargs):
        self._runtime._metadb.update_model(model_id=model_id, **kwargs)

    def get(self, model_id: int):
        return self._runtime._metadb.get_model(model_id=model_id)

    def list(self, page: int = 0, page_size: int = 10):
        return self._runtime._metadb.list_models(page=page, page_size=page_size)

    def delete(self, model_id: int):
        self._runtime._metadb.delete_model(model_id=model_id)



import uuid

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from alphatrion.metadata.base import MetaStore
from alphatrion.metadata.sql_models import (
    Base,
    Experiment,
    Metric,
    Model,
    Project,
    Run,
    Trial,
    TrialStatus,
)


# SQL-like metadata implementation, it could be SQLite, PostgreSQL, MySQL, etc.
class SQLStore(MetaStore):
    def __init__(self, db_url: str, init_tables: bool = False):
        self._engine = create_engine(db_url)
        self._session = sessionmaker(bind=self._engine)
        if init_tables:
            # create tables if not exist, will not affect existing tables.
            # Mostly used in tests.
            Base.metadata.create_all(self._engine)

    def get_project(self, project_id: str) -> Project | None:
        session = self._session()
        project = (
            session.query(Project)
            .filter(Project.uuid == project_id, Project.is_del == 0)
            .first()
        )
        session.close()
        return project

    def create_exp(
        self,
        name: str,
        project_id: str,
        description: str | None = None,
        meta: dict | None = None,
    ) -> int:
        session = self._session()
        new_exp = Experiment(
            name=name,
            project_id=project_id,
            description=description,
            meta=meta,
        )
        session.add(new_exp)
        session.commit()

        exp_id = new_exp.uuid
        session.close()

        return exp_id

    # Soft delete the experiment now. In the future, we may implement hard delete.
    def delete_exp(self, exp_id: int):
        session = self._session()
        exp = (
            session.query(Experiment)
            .filter(Experiment.uuid == exp_id, Experiment.is_del == 0)
            .first()
        )
        if exp:
            exp.is_del = 1
            session.commit()
        session.close()

    # We don't support append-only update, the complete fields should be provided.
    def update_exp(self, exp_id: int, **kwargs):
        session = self._session()
        exp = (
            session.query(Experiment)
            .filter(Experiment.uuid == exp_id, Experiment.is_del == 0)
            .first()
        )
        if exp:
            for key, value in kwargs.items():
                setattr(exp, key, value)
            session.commit()
        session.close()

    # get_exp will ignore the deleted experiments.
    def get_exp(self, exp_id: int) -> Experiment | None:
        session = self._session()
        exp = (
            session.query(Experiment)
            .filter(Experiment.uuid == exp_id, Experiment.is_del == 0)
            .first()
        )
        session.close()
        return exp

    def get_exp_by_name(self, name: str, project_id: str) -> Experiment | None:
        session = self._session()
        exp = (
            session.query(Experiment)
            .filter(
                Experiment.name == name,
                Experiment.project_id == project_id,
                Experiment.is_del == 0,
            )
            .first()
        )
        session.close()
        return exp

    # paginate the experiments in case of too many experiments.
    def list_exps(self, project_id: str, page: int, page_size: int) -> list[Experiment]:
        session = self._session()
        exps = (
            session.query(Experiment)
            .filter(Experiment.project_id == project_id, Experiment.is_del == 0)
            .offset(page * page_size)
            .limit(page_size)
            .all()
        )
        session.close()
        return exps

    def create_model(
        self,
        name: str,
        project_id: str,
        version: str = "latest",
        description: str | None = None,
        meta: dict | None = None,
    ):
        session = self._session()
        new_model = Model(
            name=name,
            project_id=project_id,
            version=version,
            description=description,
            meta=meta,
        )
        session.add(new_model)
        session.commit()
        model_id = new_model.uuid
        session.close()

        return model_id

    def update_model(self, model_id: int, **kwargs):
        session = self._session()
        model = (
            session.query(Model)
            .filter(Model.uuid == model_id, Model.is_del == 0)
            .first()
        )
        if model:
            for key, value in kwargs.items():
                setattr(model, key, value)
            session.commit()
        session.close()

    def get_model(self, model_id: int) -> Model | None:
        session = self._session()
        model = (
            session.query(Model)
            .filter(Model.uuid == model_id, Model.is_del == 0)
            .first()
        )
        session.close()
        return model

    def list_models(self, page: int, page_size: int) -> list[Model]:
        session = self._session()
        models = session.query(Model).offset(page * page_size).limit(page_size).all()
        session.close()
        return models

    def delete_model(self, model_id: int):
        session = self._session()
        model = (
            session.query(Model)
            .filter(Model.uuid == model_id, Model.is_del == 0)
            .first()
        )
        if model:
            model.is_del = 1
            session.commit()
        session.close()

    def create_trial(
        self,
        project_id: uuid.UUID,
        exp_id: uuid.UUID,
        name: str,
        description: str | None = None,
        meta: dict | None = None,
        params: dict | None = None,
        status: TrialStatus = TrialStatus.PENDING,
    ) -> uuid.UUID:
        session = self._session()
        new_trial = Trial(
            project_id=project_id,
            experiment_id=exp_id,
            name=name,
            description=description,
            meta=meta,
            params=params,
            status=status,
        )
        session.add(new_trial)
        session.commit()

        trial_id = new_trial.uuid
        session.close()

        return trial_id

    def get_trial(self, trial_id: uuid.UUID) -> Trial | None:
        session = self._session()
        trial = session.query(Trial).filter(Trial.uuid == trial_id).first()
        session.close()
        return trial

    # TODO: should we use join to get the trial by experiment name?
    def get_trial_by_name(self, trial_name: str, exp_id: uuid.UUID) -> Trial | None:
        # make sure the experiment exists
        exp = self.get_exp(exp_id)
        if exp is None:
            return None

        session = self._session()
        trial = (
            session.query(Trial)
            .filter(Trial.name == trial_name, Trial.experiment_id == exp_id)
            .first()
        )
        session.close()
        return trial

    def update_trial(self, trial_id: uuid.UUID, **kwargs):
        session = self._session()
        trial = session.query(Trial).filter(Trial.uuid == trial_id).first()
        if trial:
            for key, value in kwargs.items():
                setattr(trial, key, value)
            session.commit()
        session.close()

    def create_run(self, project_id: uuid.UUID, trial_id: uuid.UUID) -> uuid.UUID:
        session = self._session()
        new_run = Run(
            project_id=project_id,
            trial_id=trial_id,
        )
        session.add(new_run)
        session.commit()
        run_id = new_run.uuid
        session.close()
        return run_id

    def create_metric(
        self,
        project_id: uuid.UUID,
        trial_id: uuid.UUID,
        run_id: uuid.UUID,
        key: str,
        value: float,
        step: int,
    ):
        session = self._session()
        new_metric = Metric(
            project_id=project_id,
            trial_id=trial_id,
            run_id=run_id,
            key=key,
            value=value,
            step=step,
        )
        session.add(new_metric)
        session.commit()
        session.close()

    def list_metrics(self, trial_id: uuid.UUID) -> list[Metric]:
        session = self._session()
        metrics = session.query(Metric).filter(Metric.trial_id == trial_id).all()
        session.close()
        return metrics



from alphatrion.experiment.base import Experiment
from alphatrion.trial.trial import Trial, TrialConfig


class CraftExperiment(Experiment):
    """
    Craft experiment implementation.

    This experiment class offers methods to manage the experiment lifecycle flexibly.
    Opposite to other experiment classes, you need to call all these methods yourself.
    """

    def __init__(self):
        super().__init__()

    @classmethod
    def setup(
        cls,
        name: str,
        description: str | None = None,
        meta: dict | None = None,
    ) -> "CraftExperiment":
        """
        Setup the experiment. If the name already exists in the same project,
        it will refer to the existing experiment instead of creating a new one.
        """

        exp = CraftExperiment()
        exp_obj = exp._get_by_name(name=name, project_id=exp._runtime._project_id)

        # If experiment with the same name exists in the project, use it.
        if exp_obj:
            exp._id = exp_obj.uuid
        else:
            exp._create(
                name=name,
                description=description,
                meta=meta,
            )

        return exp

    def start_trial(
        self,
        name: str,
        description: str | None = None,
        meta: dict | None = None,
        params: dict | None = None,
        config: TrialConfig | None = None,
    ) -> Trial:
        """
        start_trial starts a new trial in this experiment.
        You need to call trial.cancel() to stop the trial for proper cleanup,
        unless it's a timeout trial.
        Or you can use 'async with exp.start_trial(...) as trial', which will
        automatically stop the trial at the end of the context.

        :params description: the description of the trial
        :params meta: the metadata of the trial
        :params config: the configuration of the trial

        :return: the Trial instance
        """

        trial = Trial(exp_id=self._id, config=config)
        trial._start(name=name, description=description, meta=meta, params=params)
        self.register_trial(id=trial.id, instance=trial)
        return trial



import inspect
import uuid
from functools import wraps

from opentelemetry.sdk.trace.export import ConsoleSpanExporter
from opentelemetry.semconv_ai import TraceloopSpanKindValues
from traceloop.sdk import Traceloop
from traceloop.sdk.decorators import task as _task
from traceloop.sdk.decorators import workflow as _workflow

from alphatrion.run.run import current_run_id

Traceloop.init(
    app_name="alphatrion",
    # TODO: make this configurable
    exporter=ConsoleSpanExporter(),
    disable_batch=True,
    telemetry_enabled=False,
)


def task(
    version: int | None = None,
    method_name: str | None = None,
    tlp_span_kind: TraceloopSpanKindValues | None = TraceloopSpanKindValues.TASK,
):
    return _task(
        version=version,
        method_name=method_name,
        tlp_span_kind=tlp_span_kind,
    )


def workflow(
    run_id: uuid.UUID | None = None,
    version: int | None = None,
    method_name: str | None = None,
    tlp_span_kind: TraceloopSpanKindValues | None = TraceloopSpanKindValues.WORKFLOW,
):
    """Workflow decorator with default run_id from context var.
    :param run_id: The run ID to use for the workflow as the identify name.
                   If None, use the current run ID from context var, only for tests.
    """

    def decorator(func):
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            actual_run_id = run_id or current_run_id.get()
            wrapped_func = _workflow(
                name=str(actual_run_id),
                version=version,
                method_name=method_name,
                tlp_span_kind=tlp_span_kind,
            )(func)
            return await wrapped_func(*args, **kwargs)

        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            actual_run_id = run_id or current_run_id.get()
            wrapped_func = _workflow(
                name=str(actual_run_id),
                version=version,
                method_name=method_name,
                tlp_span_kind=tlp_span_kind,
            )(func)
            return wrapped_func(*args, **kwargs)

        if inspect.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper

    return decorator



from alphatrion.run.run import current_run_id
from alphatrion.runtime.runtime import global_runtime
from alphatrion.trial.trial import current_trial_id
from alphatrion.utils import time as utime


async def log_artifact(
    paths: str | list[str],
    version: str = "latest",
):
    """
    Log artifacts (files) to the artifact registry.

    :param exp_id: the experiment ID
    :param paths: list of file paths to log.
        Support one or multiple files or a folder.
        If a folder is provided, all files in the folder will be logged.
        Don't support nested folders currently, only files in the first level
        of the folder will be logged.
    :param version: the version (tag) to log the files
    """

    if not paths:
        raise ValueError("no files specified to log")

    runtime = global_runtime()
    if runtime is None:
        raise RuntimeError("Runtime is not initialized. Please call init() first.")

    # We use experiment ID as the repo name rather than the experiment name,
    # because experiment name is not unique
    exp = runtime.current_exp
    if exp is None:
        raise RuntimeError("No running experiment found in the current context.")

    runtime._artifact.push(repo_name=str(exp.id), paths=paths, version=version)


# log_params is used to save a set of parameters, which is a dict of key-value pairs.
# should be called after starting a trial.
async def log_params(params: dict):
    trial_id = current_trial_id.get()
    if trial_id is None:
        raise RuntimeError("log_params must be called inside a Trial.")
    runtime = global_runtime()
    # TODO: should we upload to the artifact as well?
    # current_trial_id is protect by contextvar, so it's safe to use in async
    runtime._metadb.update_trial(
        trial_id=trial_id,
        params=params,
    )


# log_metrics is used to log a set of metrics at once,
# metric key must be string, value must be float.
# If save_on_best is enabled in the trial config, and the metric is the best metric
# so far, the trial will checkpoint the current data.
#
# Note: log_metrics can only be called inside a Run, because it needs a run_id.
async def log_metrics(metrics: dict[str, float]):
    run_id = current_run_id.get()
    if run_id is None:
        raise RuntimeError("log_metrics must be called inside a Run.")

    runtime = global_runtime()
    exp = runtime.current_exp

    trial_id = current_trial_id.get()
    if trial_id is None:
        raise RuntimeError("log_metrics must be called inside a Trial.")

    trial = exp.get_trial(id=trial_id)
    if trial is None:
        raise RuntimeError(f"Trial {trial_id} not found in the database.")

    # track if any metric is the best metric
    should_checkpoint = False
    should_early_stop = False
    step = trial.increment_step()
    for key, value in metrics.items():
        runtime._metadb.create_metric(
            key=key,
            value=value,
            project_id=runtime._project_id,
            trial_id=trial_id,
            run_id=run_id,
            step=step,
        )

        # TODO: should we save the checkpoint path for the best metric?
        # Always call the should_checkpoint_on_best first because
        # it also updates the best metric.
        should_checkpoint |= trial.should_checkpoint_on_best(
            metric_key=key, metric_value=value
        )
        should_early_stop |= trial.should_early_stop(metric_key=key, metric_value=value)

    if should_checkpoint:
        await log_artifact(
            paths=trial.config().checkpoint.path,
            version=utime.now_2_hash(),
        )

    if should_early_stop:
        trial.cancel()



import asyncio
from collections.abc import Callable


# Inspired by golang context package
class Context:
    def __init__(self, cancel_func: Callable | None = None, timeout=None):
        """A context for managing cancellation and timeouts.
        :param cancel_func: A function to call when the context is cancelled.
        :param timeout: Timeout in seconds. If None, no timeout is set.
        """
        self._cancel_event = asyncio.Event()
        self._cancel_func = cancel_func
        self._timeout = timeout

    def start(self):
        # If timeout is None, it means no timeout is set.
        # If timeout is negative, it means already timed out.
        if self._timeout is not None:
            asyncio.create_task(self._auto_cancel(self._timeout))

    async def _auto_cancel(self, timeout):
        await asyncio.sleep(timeout)
        self.cancel()

    def cancel(self):
        if self.cancelled():
            return
        if self._cancel_func:
            self._cancel_func()
        self._cancel_event.set()

    def cancelled(self):
        return self._cancel_event.is_set()

    async def wait(self):
        await self._cancel_event.wait()



import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass

from alphatrion.runtime.runtime import global_runtime
from alphatrion.trial import trial


@dataclass
class Experiment(ABC):
    """
    Base Experiment class. One instance one experiment, multiple trials.
    """

    __slots__ = ("_runtime", "_id", "_trials")

    def __init__(self):
        self._runtime = global_runtime()
        # All trials in this experiment, key is trial_id, value is Trial instance.
        self._trials = dict()
        self._runtime.current_exp = self

    @property
    def id(self):
        return self._id

    def get_trial(self, id: int) -> trial.Trial | None:
        return self._trials.get(id)

    async def __aenter__(self):
        if self._id is None:
            raise RuntimeError("Experiment is not set. Did you call start()?")

        exp = self._get()
        if exp is None:
            raise RuntimeError(f"Experiment {self._id} not found in the database.")

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._trials = dict()
        self._runtime.current_exp = None

    @classmethod
    @abstractmethod
    def setup(
        cls, name: str, description: str | None = None, meta: dict | None = None
    ) -> "Experiment":
        """Return a new experiment."""
        ...

    def register_trial(self, id: uuid.UUID, instance: trial.Trial):
        self._trials[id] = instance

    def unregister_trial(self, id: uuid.UUID):
        self._trials.pop(id, None)

    def _create(
        self,
        name: str,
        description: str | None = None,
        meta: dict | None = None,
    ):
        """
        :param name: the name of the experiment.
        :param description: the description of the experiment
        :param meta: the metadata of the experiment

        :return: the experiment ID
        """

        self._id = self._runtime._metadb.create_exp(
            name=name,
            description=description,
            project_id=self._runtime._project_id,
            meta=meta,
        )
        return self._id

    def _get(self):
        return self._runtime._metadb.get_exp(exp_id=self._id)

    def _get_by_name(self, name: str, project_id: str):
        return self._runtime._metadb.get_exp_by_name(name=name, project_id=project_id)

    def delete(self):
        exp = self._get()
        if exp is None:
            return

        self._runtime._metadb.delete_exp(exp_id=self._exp.id)
        # TODO: Should we make this optional as a parameter?
        tags = self._runtime._artifact.list_versions(repo_name=str(self._exp.id))
        self._runtime._artifact.delete(experiment_name=exp.name, versions=tags)



import os

import oras.client

from alphatrion import consts

SUCCESS_CODE = 201


class Artifact:
    def __init__(self, project_id: str, insecure: bool = False):
        self._project_id = project_id
        self._url = os.environ.get(consts.ARTIFACT_REGISTRY_URL)
        self._url = self._url.replace("https://", "").replace("http://", "")
        self._client = oras.client.OrasClient(
            hostname=self._url.strip("/"), auth_backend="token", insecure=insecure
        )

    def push(
        self,
        repo_name: str,
        paths: str | list[str],
        version: str = "latest",
    ):
        """
        Push files or all files in a folder to the artifact registry.
        You can specify either files or folder, but not both.
        If both are specified, a ValueError will be raised.

        :param repo_name: the name of the repository to push to
        :param paths: list of file paths or a folder path to push.
        :param version: the version (tag) to push the files under
        """

        if paths is None or not paths:
            raise ValueError("no files specified to push")

        if isinstance(paths, str):
            if os.path.isdir(paths):
                os.chdir(paths)
                files_to_push = [f for f in os.listdir(".") if os.path.isfile(f)]
            else:
                files_to_push = [paths]
        else:
            files_to_push = paths

        if not files_to_push:
            raise ValueError("No files to push.")

        url = self._url if self._url.endswith("/") else f"{self._url}/"
        target = f"{url}{self._project_id}/{repo_name}:{version}"

        try:
            self._client.push(target, files=files_to_push)
        except Exception as e:
            raise RuntimeError("Failed to push artifacts") from e

    # TODO: should we store it in the metadb instead?
    def list_versions(self, repo_name: str) -> list[str]:
        url = self._url if self._url.endswith("/") else f"{self._url}/"
        target = f"{url}{self._project_id}/{repo_name}"
        try:
            tags = self._client.get_tags(target)
            return tags
        except Exception as e:
            raise RuntimeError("Failed to list artifacts versions") from e

    def delete(self, repo_name: str, versions: str | list[str]):
        url = self._url if self._url.endswith("/") else f"{self._url}/"
        target = f"{url}{self._project_id}/{repo_name}"

        try:
            self._client.delete_tags(target, tags=versions)
        except Exception as e:
            raise RuntimeError("Failed to delete artifact versions") from e



import enum
import uuid
from datetime import UTC, datetime

from sqlalchemy import JSON, Column, DateTime, Enum, Float, Integer, String
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import declarative_base

Base = declarative_base()


class TrialStatus(enum.Enum):
    PENDING = "pending"
    RUNNING = "running"
    FINISHED = "finished"
    FAILED = "failed"


COMPLETED_STATUS = [TrialStatus.FINISHED, TrialStatus.FAILED]


class Project(Base):
    __tablename__ = "projects"

    uuid = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column(String, nullable=False)
    description = Column(String, nullable=True)

    created_at = Column(DateTime(timezone=True), default=datetime.now(UTC))
    updated_at = Column(
        DateTime(timezone=True), default=datetime.now(UTC), onupdate=datetime.now(UTC)
    )
    is_del = Column(Integer, default=0, comment="0 for not deleted, 1 for deleted")


# Define the Experiment model for SQLAlchemy
class Experiment(Base):
    __tablename__ = "experiments"

    uuid = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column(String, nullable=False)
    description = Column(String, nullable=True)
    project_id = Column(UUID(as_uuid=True), nullable=False)
    meta = Column(JSON, nullable=True, comment="Additional metadata for the experiment")

    created_at = Column(DateTime(timezone=True), default=datetime.now(UTC))
    updated_at = Column(
        DateTime(timezone=True), default=datetime.now(UTC), onupdate=datetime.now(UTC)
    )
    is_del = Column(Integer, default=0, comment="0 for not deleted, 1 for deleted")


class Trial(Base):
    __tablename__ = "trials"

    uuid = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    project_id = Column(UUID(as_uuid=True), nullable=False)
    experiment_id = Column(UUID(as_uuid=True), nullable=False)
    name = Column(String, nullable=False)
    description = Column(String, nullable=True)
    meta = Column(JSON, nullable=True, comment="Additional metadata for the trial")
    params = Column(JSON, nullable=True, comment="Parameters for the experiment")
    status = Column(
        Enum(TrialStatus),
        default=TrialStatus.PENDING,
        nullable=False,
        comment="Status of the trial",
    )

    created_at = Column(DateTime(timezone=True), default=datetime.now(UTC))
    updated_at = Column(
        DateTime(timezone=True), default=datetime.now(UTC), onupdate=datetime.now(UTC)
    )


class Run(Base):
    __tablename__ = "runs"

    uuid = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    project_id = Column(UUID(as_uuid=True), nullable=False)
    trial_id = Column(UUID(as_uuid=True), nullable=False)

    created_at = Column(DateTime(timezone=True), default=datetime.now(UTC))
    updated_at = Column(
        DateTime(timezone=True), default=datetime.now(UTC), onupdate=datetime.now(UTC)
    )


class Model(Base):
    __tablename__ = "models"

    uuid = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column(String, nullable=False, unique=True)
    description = Column(String, nullable=True)
    project_id = Column(UUID(as_uuid=True), nullable=False)
    version = Column(String, nullable=False)
    meta = Column(JSON, nullable=True, comment="Additional metadata for the model")

    created_at = Column(DateTime(timezone=True), default=datetime.now(UTC))
    updated_at = Column(
        DateTime(timezone=True), default=datetime.now(UTC), onupdate=datetime.now(UTC)
    )
    is_del = Column(Integer, default=0, comment="0 for not deleted, 1 for deleted")


class Metric(Base):
    __tablename__ = "metrics"

    uuid = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    key = Column(String, nullable=False)
    value = Column(Float, nullable=False)
    project_id = Column(UUID(as_uuid=True), nullable=False)
    trial_id = Column(UUID(as_uuid=True), nullable=False)
    run_id = Column(UUID(as_uuid=True), nullable=False)
    step = Column(Integer, nullable=False, default=0)
    created_at = Column(DateTime(timezone=True), default=datetime.now(UTC))



from alphatrion.experiment.craft_exp import CraftExperiment
from alphatrion.log.log import log_artifact, log_metrics, log_params
from alphatrion.runtime.runtime import init
from alphatrion.tracing.tracing import task, workflow
from alphatrion.trial.trial import CheckpointConfig, TrialConfig

__all__ = [
    "log_artifact",
    "log_params",
    "log_metrics",
    "CraftExperiment",
    "init",
    "TrialConfig",
    "CheckpointConfig",
    "task",
    "workflow",
]



import hashlib
from datetime import UTC, datetime


# ruff: noqa: PLR2004
def humanize_time(timestamp: str) -> str:
    creation_time = datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%SZ").replace(
        tzinfo=UTC
    )
    t = datetime.now(UTC) - creation_time

    if t.days > 0:
        age = f"{t.days}d"
    elif t.seconds >= 3600:
        age = f"{t.seconds // 3600}h"
    elif t.seconds >= 60:
        age = f"{t.seconds // 60}m"
    else:
        age = f"{t.seconds}s"

    return age


def now_2_hash() -> str:
    timestamp = str(int(datetime.now(UTC).timestamp()))
    unique_hash = hashlib.sha1(timestamp.encode()).hexdigest()[:7]

    print("Generated hash from timestamp:", timestamp, "->", unique_hash)
    return unique_hash
